module Lib
    ( PopulationStats(..)
    , TspResult(..)
    , toJsonStr
    , createCities
    , initialPopulation
    , runGA
    , calcPathLength
    ) where

import Control.DeepSeq (deepseq, NFData, rnf)
import Control.Monad.ST
import Control.Parallel.Strategies (using, parList, rdeepseq)
import Data.List (sortOn)
import qualified Data.IntSet as Set (fromList, notMember)
import Data.Vector.Unboxed (unsafeFreeze, unsafeThaw, (!))
import qualified Data.Vector.Unboxed as V (fromList, toList)
import Data.Vector.Unboxed.Mutable (MVector, swap)
import List.Shuffle (shuffle)
import System.Random (randomR, split, StdGen)

type Index = Int
type Path = [Index]
type Location = (Double, Double)

data Cities = Cities
  { cCount :: Int
  , cCoords :: [Location]
  , cDistMap :: (Index -> Index -> Double)
  }

data Population = Population
  { pPopSize :: Int
  , pPathSize :: Int
  , pPaths :: [Path]
  , pPathLengths :: [Double]
  , pScores :: [Double]
  , pTotalScore :: Double
  }

mkPopulation :: Int -> Int -> [Path] -> [Double] -> Population
mkPopulation popSize pathSize paths pathLengths = Population popSize pathSize paths pathLengths scores (sum scores)
  where scores = map fitness pathLengths

instance NFData Population where
  rnf (Population popSize pathSize paths pathLengths scores totalScore) =
    rnf popSize `seq`
    rnf pathSize `seq`
    rnf paths `seq`
    rnf pathLengths `seq`
    rnf scores `seq`
    rnf totalScore

data PopulationStats = PopulationStats
  { sBestPath :: Path
  , sMinPathLength :: Double
  , sMaxPathLength :: Double
  , sAvgPathLength :: Double
  }

mkPopulationStats :: Population -> PopulationStats
mkPopulationStats pop = PopulationStats bestPath minPathLength maxPathLength avgPathLength
  where minPathLength = minimum $ pPathLengths pop
        (maxPathLength, bestPath) = maximum $ zip (pPathLengths pop) (pPaths pop)
        avgPathLength = (sum $ pPathLengths pop) / (fromIntegral $ pPopSize pop)

data TspResult = TspResult
  { tspCities :: Cities
  , tspFinalStats :: PopulationStats
  }

toJsonStr :: TspResult -> String
toJsonStr (TspResult cities (PopulationStats bestPath _ bestPathLength _)) =
    "{\n  \"cities\": " ++ show [[x,y] | (x,y) <- cCoords cities] ++
    ",\n  \"pathIndices\": " ++ show bestPath ++
    ",\n  \"score\": " ++ show bestPathLength ++
    "\n}"

createCities :: Int -> StdGen -> (Cities, StdGen)
createCities n r = (Cities n coords distMap, last (r:rs))
  where (coords, rs) = unzip $ take n $ drop 1 $ iterate randomCoord ((-1,-1), r)
        randomCoord (_, r') = randomR ((0,0), (400,400)) r'
        distMap = mkDistanceMap n coords

mkDistanceMap :: Int -> [Location] -> (Index -> Index -> Double)
mkDistanceMap n coords = distanceMap
  where distances = V.fromList [
                      sqrt $ (x1-x0)*(x1-x0) + (y1-y0)*(y1-y0) |
                      (x0,y0) <- coords, (x1,y1) <- coords
                    ]
        distanceMap i j = distances ! (i*n + j)

initialPopulation :: Cities -> Int -> (Path -> Double) -> StdGen -> (Population, StdGen)
initialPopulation _ 0 _ r = (mkPopulation 0 0 [] [], r)
initialPopulation cities popSize pathLength r = (mkPopulation popSize pathSize paths pathLengths, last rs)
  where pathSize    = cCount cities
        (paths, rs) = unzip $ take popSize $ drop 1 $ iterate (uncurry shuffle) ([0..pathSize-1], r)
        pathLengths = map pathLength paths

nextGeneration :: Population -> (Path -> Double) -> StdGen -> (Population, StdGen)
nextGeneration pop pathLength r = (combinedPop, r')
  where popSize = pPopSize pop
        (newPop, r') = nextGeneration' pop (popSize-20) pathLength r
        bestPop = bestPaths pop 20
        paths = pPaths newPop ++ pPaths bestPop
        pathLengths = pPathLengths newPop ++ pPathLengths bestPop
        scores = pScores newPop ++ pScores bestPop
        totalScore = pTotalScore newPop + pTotalScore bestPop
        combinedPop = pop { pPaths = paths, pPathLengths = pathLengths, pScores = scores, pTotalScore = totalScore }

nextGeneration' :: Population -> Int -> (Path -> Double) -> StdGen -> (Population, StdGen)
nextGeneration' _ 0 _ r = (mkPopulation 0 0 [] [], r)
nextGeneration' pop n pathLength r = (mkPopulation n (pPathSize pop) paths pathLengths, last rs)
  where (paths, rs) = unzip $ take n $ drop 1 $ iterate (\(_, r') -> spawnPath pop r') ([], r)
        pathLengths = map pathLength paths

runNGenerations :: Population -> Int -> (Path -> Double) -> StdGen -> (Population, StdGen)
runNGenerations pop 0 _ r = (pop, r)
runNGenerations pop n pathLength r = pop' `deepseq` runNGenerations pop' (n-1) pathLength r'
  where (pop', r') = nextGeneration pop pathLength r

runKEvolutionaryBranches :: Population -> Int -> Int -> (Path -> Double) -> StdGen -> (Population, StdGen)
runKEvolutionaryBranches pop 0 _ _ r = (pop, r)
runKEvolutionaryBranches pop k n pathLength r = (pop', head rs')
  where rs = take k $ iterate (fst . split) r
        (pops, rs') = unzip (map (runNGenerations pop n pathLength) rs `using` parList rdeepseq)
        pop' = mergePopulations pops

runGA :: Population -> Int -> Int -> Int -> (Path -> Double) -> StdGen -> [PopulationStats]
runGA pop m k n pathLength r = popStats
  where popStats = map mkPopulationStats
                 $ fst $ unzip $ take (m + 1)
                 $ iterate (\(pop', r') -> runKEvolutionaryBranches pop' k n pathLength r') (pop, r)

mergePopulations :: [Population] -> Population
mergePopulations pops = bestPaths (Population (-1) pathSize paths pathLengths scores (-1)) popSize
  where popSize = pPopSize $ head pops
        pathSize = pPathSize $ head pops
        paths = concatMap pPaths pops
        pathLengths = concatMap pPathLengths pops
        scores = concatMap pScores pops

bestPaths :: Population -> Int -> Population
bestPaths (Population _ pathSize paths pathLengths scores _) n = Population n pathSize paths' pathLengths' scores' (sum scores')
  where (paths', pathLengths', scores') = unzip3 $ take n $ sortOn (\(_, _, score) -> negate score)
                                        $ zip3 paths pathLengths scores

spawnPath :: Population -> StdGen -> (Path, StdGen)
spawnPath pop r = mutate child pathSize r'''
  where pathSize = pPathSize pop
        (parent1, r') = pickOne pop r
        (parent2, r'') = pickOne pop r'
        (child, r''') = crossOver parent1 parent2 pathSize r''

fitness :: Double -> Double
fitness pathLen = 1 / (1 + (pathLen ** 16))

calcPathLength :: Cities -> Path -> Double
calcPathLength _ [] = 0
calcPathLength _ [_] = 0
calcPathLength cities path = pathLength' 0 path
  where distanceMap = cDistMap cities
        pathLength' acc [] = acc
        pathLength' acc [i] = acc `seq` acc + distanceMap i (head path)
        pathLength' acc (i:j:path') = acc' `seq` pathLength' acc' (j:path')
          where acc' = acc + distanceMap i j

pickOne :: Population -> StdGen -> (Path, StdGen)
pickOne (Population _ _ paths _ scores totalScore) r = (pickOne' paths scores 0, r')
  where (randVal, r') = randomR (0, totalScore) r
        pickOne' [] _ _ = []
        pickOne' _ [] _ = []
        pickOne' (path:paths') (score:scores') accScore = if accScore + score >= randVal
                                                          then path
                                                          else pickOne' paths' scores' (accScore + score)

crossOver :: Path -> Path -> Int -> StdGen -> (Path, StdGen)
crossOver p1 p2 pathSize r = (newPath, r')
  where (p1', r') = randomSublist p1 pathSize r
        p1'' = Set.fromList p1'
        p2' = filter (`Set.notMember` p1'') p2
        newPath = p1' ++ p2'

randomSublist :: Path -> Int -> StdGen -> (Path, StdGen)
randomSublist path pathSize r = (take (j-i+1) $ drop i path, r'')
  where (i, r') = randomR (0, pathSize - 1) r
        (j, r'') = randomR (i, pathSize - 1) r'

mutate :: Path -> Int -> StdGen -> (Path, StdGen)
mutate path pathSize r = runST $ do
  path' <- unsafeThaw (V.fromList path)
  r' <- mutateInPlace path' 0 pathSize r
  path'' <- unsafeFreeze path'
  return (V.toList path'', r')

mutateInPlace :: (MVector s Int) -> Int -> Int -> StdGen -> ST s StdGen
mutateInPlace _ i pathSize r | i >= pathSize - 1 = do return r
mutateInPlace p i pathSize r = do
  let (randVal, r') = randomR (0 :: Double, 1) r
      j = if randVal < 0.01 then (i+1) `mod` pathSize else i
  swap p i j
  mutateInPlace p (i+1) pathSize r'
